@@ -1,8 +1,8 @@
-use std::{error, fmt};
+use std::{borrow, error, fmt};
 
 #[derive(Debug)]
-pub struct FormattedError {
-    pub message: String,
+struct FormattedError {
+    message: borrow::Cow<'static, str>,
 }
 
 impl fmt::Display for FormattedError {
@@ -13,10 +13,22 @@ impl fmt::Display for FormattedError {
 
 impl error::Error for FormattedError {}
 
+#[inline]
+#[must_use]
+pub fn error_from_args(args: fmt::Arguments<'_>) -> impl error::Error + Send + Sync + 'static {
+    FormattedError {
+        message: if let Some(message) = args.as_str() {
+            borrow::Cow::Borrowed(message)
+        } else {
+            borrow::Cow::Owned(fmt::format(args))
+        },
+    }
+}
+
 #[derive(Debug)]
-pub struct FormattedWrapError {
-    pub message: String,
-    pub source: crate::Error,
+struct FormattedWrapError {
+    message: borrow::Cow<'static, str>,
+    source: crate::Error,
 }
 
 impl fmt::Display for FormattedWrapError {
@@ -31,6 +43,23 @@ impl error::Error for FormattedWrapError {
     }
 }
 
+#[inline]
+#[must_use]
+pub fn wrap_error_from_args(
+    source: impl Into<crate::Error>,
+    args: fmt::Arguments<'_>,
+) -> impl error::Error + Send + Sync + 'static {
+    let message = if let Some(message) = args.as_str() {
+        borrow::Cow::Borrowed(message)
+    } else {
+        borrow::Cow::Owned(fmt::format(args))
+    };
+    FormattedWrapError {
+        source: source.into(),
+        message,
+    }
+}
+
 pub(crate) struct WrapError {
     pub(crate) inner: crate::Error,
 }
@@ -52,9 +81,3 @@ impl error::Error for WrapError {
         self.inner.source()
     }
 }
-
-pub fn make_opaque(
-    error: impl error::Error + Send + Sync + 'static,
-) -> impl error::Error + Send + Sync + 'static {
-    error
-}
